/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Copyright (C) 2005-2008 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */
package weka4ws.core;

/** 
 * @author Marco Lackovic 
 */
public class DatasetFile {

    private String fileName;
    private long crc;
    private int processes = 1;
    private boolean beingCompressed = true;

    public boolean isBeingCompressed() {
        return beingCompressed;
    }

    public void setBeingCompressed(boolean beeingCompressed) {
        this.beingCompressed = beeingCompressed;
    }

    public DatasetFile(String fileName2, long crc2) {
        this.fileName = fileName2;
        this.crc = crc2;
    }

    public synchronized int getProcesses() {
        return processes;
    }

    public synchronized void addProcess() {
        this.processes++;
    }

    public synchronized void removeProcess() {
        this.processes--;
    }

    public synchronized boolean equals(DatasetFile d) {
        if (this.fileName.equals(d.fileName) && (this.crc == d.crc)) {
            return true;
        }
        return false;
    }

    /**
     * Returns whether a process is the last one, if not decreases the number
     * of processes 
     */
    public synchronized boolean checkAndRemove() {
        if (processes <= 1) {
            return true;
        }
        processes--;
        return false;
    }

}
