/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Copyright (C) 2005-2008 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */
package weka4ws.core;

import java.util.ArrayList;

import weka4ws.stubs.service.Dataset;

/** 
 * @author Marco Lackovic 
 */
public class FileManager {

    protected static ArrayList files = new ArrayList();

    public synchronized boolean allowedToCompress(DatasetFile df) {
        int size = files.size();
        if (size > 0) {
            int i = 0;
            boolean found = false;
            DatasetFile df2 = (DatasetFile) files.get(0);
            while (i < size && !found) {
                df2 = (DatasetFile) files.get(i);
                if (df2.equals(df)) {
                    found = true;
                    df2.addProcess();
                }
                i++;
            }
            if (found) {
                try {
                    while (df2.isBeingCompressed()) {
                        wait();
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                return false;
            } else {
                files.add(df);
                return true;
            }
        } else {
            files.add(df);
            return true;
        }
    }

    public synchronized void fileCompressed(DatasetFile df) {
        int size = files.size();
        int i = 0;
        boolean found = false;
        DatasetFile df2;
        while (i < size && !found) {
            df2 = (DatasetFile) files.get(i);
            if (df2.equals(df)) {
                found = true;
                df2.setBeingCompressed(false);
                notifyAll();
            }
            i++;
        }
    }

    public synchronized boolean allowedToDelete(DatasetFile d) {
        int size = files.size();
        int i = 0;
        boolean found = false;
        DatasetFile df2;
        while (i < size && !found) {
            df2 = (DatasetFile) files.get(i);
            if (df2.equals(d)) {
                found = true;
                if (df2.checkAndRemove()) {
                    files.remove(df2);
                    return true;
                }
                return false;
            }
            i++;
        }
        return true;
    }
}
