/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  HostCheckConnectionThread.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.io.IOException;
import java.net.ConnectException;
import java.net.NoRouteToHostException;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Tries to establish a connection with the given <code>host</code> at the
 * given <code>port</code> in order to check whether the <code>service</code>
 * is available and running.
 * 
 * @author Marco Lackovic
 * @version 1.0
 * @see weka4ws.core.Hosts
 * @see weka4ws.core.HostCheckThread
 */
public class HostCheckConnectionThread extends Thread {

    private String host;

    private boolean connFailed = false;

    private int port;

    private Socket socket;

    private String serviceName;

    private String message = "";

    public HostCheckConnectionThread(String host, String port,
            String serviceName) {
        this.host = host;
        this.port = Integer.valueOf(port).intValue();
        this.serviceName = serviceName;
    }

    public void run() {
        try {
            socket = new Socket(host, port);
        } catch (ConnectException e) {
            connFailed = true;
            message = "Failed to contact " + serviceName + " at host \"" + host
                    + "\".";
        } catch (NoRouteToHostException e) {
            connFailed = true;
            message = "Host \"" + host + "\" is not reachable.";
        } catch (UnknownHostException e) {
            connFailed = true;
            message = "Host name \"" + host + "\" can't be resolved.";
        } catch (IOException e) {
            connFailed = true;
            e.printStackTrace();
        } catch (Exception e) {
            connFailed = true;
            e.printStackTrace();
        } finally {
            try {
                if (socket != null)
                    socket.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

    }

    public boolean connectionFailed() {
        return connFailed;
    }

    public String getErrorMessage() {
        return message;
    }
}
