/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  HostCheckGlobusThread.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.net.UnknownHostException;

/**
 * Tries to establish a series of connections in order to check whether a given
 * host is suitable for globus transactions.
 * 
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class HostCheckThread extends Thread {

    private String host;

    private String servicePort;

    private String gridFTPPort;

    private String logging;

    public static final int TIME_OUT = 6000;

    private String message = "";

    private boolean isFunctional = false;

    public HostCheckThread(String host, String servicePort, String gridFTPPort,
            String logging) {
        this.host = host;
        this.servicePort = servicePort;
        this.gridFTPPort = gridFTPPort;
        this.logging = logging;
    }

    public void run() {

        String service = "";

        try {
            // service = "globus container";
            // HostCheckConnectionThread checkContainer = new
            // HostCheckConnectionThread(
            // host, servicePort, service);
            // checkContainer.start();
            // checkContainer.join(TIME_OUT);
            // if (checkContainer.isAlive()) {
            // checkContainer.stop();
            // throw new HostConnectionTimedOutException();
            // } else if (checkContainer.connectionFailed()) {
            // message = checkContainer.getErrorMessage();
            // throw new Exception();
            // }
            service = "gridFTP";
            HostCheckConnectionThread checkGridFTP = new HostCheckConnectionThread(
                    host, gridFTPPort, service);
            checkGridFTP.start();
            checkGridFTP.join(TIME_OUT);
            if (checkGridFTP.isAlive()) {
                checkGridFTP.interrupt();
                throw new HostConnectionTimedOutException();
            } else if (checkGridFTP.connectionFailed()) {
                message = "Connection to " + service + " at \"" + host
                        + "\" failed.";
                throw new Exception();
            }

            service = "Weka4WS";
            HostCheckVersionThread checkWeka4WSVersion = new HostCheckVersionThread(
                    host, servicePort);
            checkWeka4WSVersion.start();
            checkWeka4WSVersion.join(TIME_OUT);
            if (checkWeka4WSVersion.isAlive()) {
                checkWeka4WSVersion.interrupt();
                throw new HostConnectionTimedOutException();
            } else if (checkWeka4WSVersion.connectionFailed()) {
                message = checkWeka4WSVersion.getErrorMessage();
                if (message.length() == 0) {
                    message = "Connection to " + service + " at \"" + host
                            + "\" failed.";
                }
                throw new Exception();
            }

            isFunctional = true;
            message = "Host \"" + host + "\" fully functional.";

        } catch (UnknownHostException e) {
            message = "Host name \"" + host + "\" can't be resolved.";
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (HostConnectionTimedOutException e) {
            message = "Connection to " + service + " at \"" + host
                    + "\" timed out.";
        } catch (Exception e) {

        }
    }

    public String getHost() {
        return host;
    }

    public String getServicePort() {
        return servicePort;
    }

    public String getGridFTPPort() {
        return gridFTPPort;
    }

    public String getLogging() {
        return logging;
    }

    public String getMessage() {
        return message;
    }

    public boolean isFunctional() {
        return isFunctional;
    }

}
