/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  HostCheckVersionThread.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import org.apache.axis.AxisFault;
import org.apache.axis.types.URI.MalformedURIException;

import weka4ws.stubs.service.VersionParameters;

/**
 * Checks whether the service version equals the client version.
 * 
 * @author Marco Lackovic
 * @version 1.0
 * @see weka4ws.core.Hosts
 * @see weka4ws.core.HostCheckThread
 */
public class HostCheckVersionThread extends Thread {

    private String host;

    private boolean connFailed = true;

    private int port;

    private String message = "";

    public HostCheckVersionThread(String host, String port) {
        this.host = host;
        this.port = Integer.valueOf(port).intValue();
    }

    public void run() {
        RemoteService weka4WSService;
        try {
            weka4WSService = new RemoteService(host, port);
            String version = (weka4WSService.getPortType())
                    .getVersion(new VersionParameters());
            if (!version.equals(WUtil.VERSION)) {
                message = "Version mismatch between Weka4WS client and service at "
                        + host;
            } else {
                connFailed = false;
            }
        } catch (AxisFault e) {
            String temp = e.getMessage();
            if (temp.indexOf("could not find a target service to invoke") > 0) {
                message = "Weka4WSService not found at \"" + host + "\"";
            } else if (temp.indexOf("GSS") > 0) {
                message = temp.substring(temp.indexOf("GSS")) + " at \"" + host
                        + "\"";
            } else if (temp.indexOf("EOF") > 0) {
                message = temp.substring(temp.indexOf("EOF")) + " at \"" + host
                        + "\"";
            } else if (temp.indexOf("Connection refused") > 0) {
                message = "Connection to Globus Container at \"" + host
                        + "\" refused";
            } else {
                e.printStackTrace();
            }
        } catch (MalformedURIException e) {
            message = "MalformedURIException";
        } catch (ServiceException e) {
            message = "ServiceException";
        } catch (RemoteException e) {
            message = "RemoteException";
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public boolean connectionFailed() {
        return connFailed;
    }

    public String getErrorMessage() {
        return message;
    }
}
