/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  NotificationReceiver.java
 *  Copyright (C) 2005 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.rmi.RemoteException;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.oasis.wsrf.properties.InvalidResourcePropertyQNameFaultType;
import org.oasis.wsrf.properties.ResourceUnknownFaultType;

import weka.gui.Logger;
import weka4ws.stubs.service.Result;
import weka4ws.stubs.service.Weka4WSPortType;

/**
 * @author Oreste Verta
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class NotificationReceiver implements NotifyCallback {

    public static final String NAMESPACE = "http://grid.deis.unical.it/weka4ws/Weka4WSService";

    public static final QName RP_RESULT = new QName(NAMESPACE, "Result");

    public static final short PUSH_MODE = 0;

    public static final short PULL_MODE = 1;

    public static final short POLLING_TIME = 10000;

    protected short notificationStyle = PULL_MODE;

    protected String host;

    protected Logger log;

    private boolean taskStopped = false;

    public NotificationReceiver(String hostAddress, Logger log)
            throws Exception {

        this.host = hostAddress;
        this.log = log;
    }

    /**
     * @return the result
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws InvalidResourcePropertyQNameFaultType
     * @throws DeserializationException
     * @uml.property name="result"
     */
    public synchronized Result getResult(Weka4WSPortType port)
            throws InvalidResourcePropertyQNameFaultType,
            ResourceUnknownFaultType, RemoteException, DeserializationException {

        GetResourcePropertyResponse resultRP;
        MessageElement msg;

        Result result = new Result();
        result.setReady(false);

        // Runtime r = Runtime.getRuntime();

        // System.out.println("totalMemory = " + r.totalMemory());

        while (!result.isReady()) {

            switch (notificationStyle) {

            case PULL_MODE: // default

                try {
                    wait(POLLING_TIME);
                } catch (Exception e) {
                    System.err.println(e);
                }

                break;

            case PUSH_MODE:

                try {
                    wait();
                } catch (Exception e) {
                    System.err.println(e);
                }

                break;
            }

            if (taskStopped) {
                return null;
            } else {
                try {
                    // High throughput can cause an OutOfMemoryError exception.
                    // Fixes will appear in the next development release.
                    resultRP = port.getResourceProperty(RP_RESULT);
                    msg = resultRP.get_any()[0];
                    result = (Result) ObjectDeserializer.toObject(msg,
                            Result.class);

                } catch (OutOfMemoryError e) {
                    e.printStackTrace();
                }
            }
        }
        return result;
    }

    public synchronized void deliver(List topicPath,
            EndpointReferenceType producer, Object message) {
        if (notificationStyle == PULL_MODE) {
            notificationStyle = PUSH_MODE;
            WUtil.logMessage(log, host,
                    "   Switching to PUSH mode notification style");
        }
        notify();
    }

    public synchronized void stop() {
        taskStopped = true;
        notify();
    }

    public void setNotificationStyle(short notificationStyle) {
        this.notificationStyle = notificationStyle;
    }

}
