/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  RemoteResourse.java
 *  Copyright (C) 2008 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.namespace.QName;
import javax.xml.rpc.ServiceException;

import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.URI.MalformedURIException;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.WSNConstants;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.ResourceNotDestroyedFaultType;
import org.oasis.wsrf.lifetime.ResourceUnknownFaultType;

import weka4ws.stubs.factory.CreateResource;
import weka4ws.stubs.factory.CreateResourceResponse;
import weka4ws.stubs.factory.Weka4WSFactoryPortType;
import weka4ws.stubs.factory.service.Weka4WSFactoryServiceAddressingLocator;
import weka4ws.stubs.service.NotifCheckParameters;
import weka4ws.stubs.service.Weka4WSPortType;
import weka4ws.stubs.service.service.Weka4WSServiceAddressingLocator;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class RemoteResourse {

    protected Weka4WSPortType port;

    protected static ArrayList ports = new ArrayList();

    protected String hostAddress = "localhost";
    protected int servicePort = 8443;
    protected String resourceName = "";

    public static final String FACTORY_SERVICE = "wsrf/services/Weka4WSFactoryService";

    public static final String NAMESPACE = "http://grid.deis.unical.it/weka4ws/Weka4WSService";

    public RemoteResourse(String hostAddress, int servicePort,
            String resourceName) throws MalformedURIException,
            ServiceException, RemoteException {
        this.hostAddress = hostAddress;
        this.servicePort = servicePort;
        this.resourceName = resourceName;

        Weka4WSFactoryServiceAddressingLocator factoryLocator = new Weka4WSFactoryServiceAddressingLocator();
        Weka4WSServiceAddressingLocator instanceLocator = new Weka4WSServiceAddressingLocator();
        String factoryURI = "https://" + hostAddress + ":" + servicePort + "/"
                + FACTORY_SERVICE;
        EndpointReferenceType factoryEPR, instanceEPR;
        Weka4WSFactoryPortType wekaFactory;

        // Get factory portType
        factoryEPR = new EndpointReferenceType();
        factoryEPR.setAddress(new Address(factoryURI));
        wekaFactory = factoryLocator.getWeka4WSFactoryPortTypePort(factoryEPR);
        CreateResourceResponse createResponse = wekaFactory
                .createResource(new CreateResource());
        instanceEPR = createResponse.getEndpointReference();

        // Get instance portType
        port = instanceLocator.getWeka4WSPortTypePort(instanceEPR);
        ports.add(port);
    }

    public Weka4WSPortType getPortType() throws Exception {
        return port;
    }

    public void notifSubscribe(NotificationReceiver receiver,
            NotificationConsumerManager consumer) throws Exception {
        /*
         * // Tell the stub to use GSI Secure Conversation with encryption
         * ((Stub) port)._setProperty(Constants.GSI_SEC_CONV,
         * Constants.ENCRYPTION); ((Stub)
         * port)._setProperty(Constants.AUTHORIZATION, NoAuthorization
         * .getInstance());
         * 
         * // Delegate credentials to the service ((Stub)
         * port)._setProperty(GSIConstants.GSI_MODE,
         * GSIConstants.GSI_MODE_FULL_DELEG);
         */
        consumer.startListening();
        EndpointReferenceType consumerEPR = consumer
                .createNotificationConsumer(receiver);

        Subscribe request = new Subscribe();
        request.setUseNotify(Boolean.TRUE);
        request.setConsumerReference(consumerEPR);

        QName RP_MODEL = new QName(NAMESPACE, resourceName);
        TopicExpressionType topicExpression = new TopicExpressionType();
        topicExpression.setDialect(WSNConstants.SIMPLE_TOPIC_DIALECT);
        topicExpression.setValue(RP_MODEL);

        request.setTopicExpression(topicExpression);
        port.subscribe(request);
        port.notifCheck(new NotifCheckParameters());
    }

    public synchronized void destroy() {
        try {
            port.destroy(new Destroy());
        } catch (ResourceNotDestroyedFaultType e) {
        } catch (ResourceUnknownFaultType e) {
        } catch (RemoteException e) {
        }
        ports.remove(port);
    }

    public static synchronized void destroyAll()
            throws ResourceNotDestroyedFaultType, ResourceUnknownFaultType,
            RemoteException {
        if (ports.size() > 0) {
            for (Iterator iterator = ports.iterator(); iterator.hasNext();) {
                Weka4WSPortType port = (Weka4WSPortType) iterator.next();
                port.destroy(new Destroy());
            }
            ports.clear();
        }
    }
}
