/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Copyright (C) 2005-2008 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.core;

import java.io.File;
import java.net.URL;
import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import org.apache.axis.types.URI.MalformedURIException;

import weka4ws.stubs.service.TestOptions;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public class RemoteTestTask extends RemoteTask {

    public RemoteTestTask(ConnectionParameters connectionParameters,
            RemoteParameters parameters) throws MalformedURIException,
            RemoteException, ServiceException {
        super(connectionParameters, parameters);
    }

    public void setTestOptions(int testMode, int percent, int numFolds) {
        TestOptions testOptions = new TestOptions();
        testOptions.setTestMode(testMode);
        testOptions.setPercent(percent);
        testOptions.setNumFolds(numFolds);
        ((RemoteTestParameters) this.parameters).setTestOptions(testOptions);
    }

    public void setTestOptions(TestOptions testOptions) {
        ((RemoteTestParameters) this.parameters).setTestOptions(testOptions);
    }

    public void setTestset(String testsetFilePath) throws Exception {
        // String localHostName = (InetAddress.getLocalHost()).getHostName();

        String filePath;
        if (testsetFilePath.startsWith("http://")) {
            URL url = new URL(testsetFilePath);
            String fileName = File.separator
                    + (new File(url.getFile())).getName();
            String currentDir = System.getProperty("user.home");
            // TODO Aggiungere una throw exception se la creazione della
            // cartella non va a buon fine
            (new File(currentDir + File.separator + TEMP_DIR)).mkdir();
            WUtil.downloadDataset(testsetFilePath, fileName, TEMP_DIR);
            filePath = currentDir + File.separator + TEMP_DIR + File.separator
                    + fileName;
        } else {
            filePath = testsetFilePath;
        }

        long crcValue = getChecksumValue(filePath);

        testset.setFilePath(filePath);
        testset.setFileName(filePath.substring(filePath
                .lastIndexOf(File.separator) + 1));
        testset.setDirPath(filePath.substring(0, filePath
                .lastIndexOf(File.separator)));
        testset.setCrc(crcValue);

        ((RemoteTestParameters) this.parameters).setTestset(testset);
    }

}
