/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Copyright (C) 2005-2008 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */
package weka4ws.core;

import java.util.ArrayList;

/** 
 * @author Marco Lackovic 
 */
public class TransfersManager {

    protected static ArrayList transfers = new ArrayList();

    public synchronized boolean getLock(Transfer t) {
        int size = transfers.size();
        if (size > 0) {
            int i = 0;
            boolean found = false;
            Transfer t2 = (Transfer) transfers.get(0);
            while (i < size && !found) {
                t2 = (Transfer) transfers.get(i);
                if (t2.equals(t)) {
                    found = true;
                    t2.addProcess();
                }
                i++;
            }
            if (found) {
                try {
                    while (t2.isInProgress()) {
                        wait();
                    }
                    if (t2.checkAndRemove()) {
                        transfers.remove(t2);
                    }
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                return false;
            } else {
                transfers.add(t);
                return true;
            }
        } else {
            transfers.add(t);
            return true;
        }
    }

    public synchronized void releaseLock(Transfer t) {
        int size = transfers.size();
        int i = 0;
        boolean found = false;
        Transfer t2;
        while (i < size && !found) {
            t2 = (Transfer) transfers.get(i);
            if (t2.equals(t)) {
                found = true;
                t2.setInProgress(false);

                if (t2.checkAndRemove()) {
                    transfers.remove(t2);
                } else {
                    notifyAll();
                }
            }
            i++;
        }
    }
}
