/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    AssociationsPanel.java
 *    Copyright (C) 1999 Eibe Frank
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2005 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.explorer;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JViewport;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import weka.associations.Associator;
import weka.core.Attribute;
import weka.core.Instances;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.gui.GenericObjectEditor;
import weka.gui.Logger;
import weka.gui.PropertyPanel;
import weka.gui.SaveBuffer;
import weka.gui.TaskLogger;
import weka4ws.core.AssociationRulesParameters;
import weka4ws.core.ButtonDisabler;
import weka4ws.core.ConnectionParameters;
import weka4ws.core.Hosts;
import weka4ws.core.RemoteAssociator;
import weka4ws.core.RemoteTask;
import weka4ws.core.TaskStoppedException;
import weka4ws.core.WUtil;
import weka4ws.stubs.service.ModelResult;
import weka4ws.stubs.service.Result;

/**
 * This panel allows the user to select, configure, and run a scheme that learns
 * associations.
 * 
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 * @author Oreste Verta
 * 
 * @version $Revision: 1.17.2.2 $
 */
public class AssociationsPanel extends ExplorerPanel {

    /* ============== Weka4WS modifications ============== */

    private static final long serialVersionUID = 1L;

    /**
     * Starts running the currently configured associator with the current
     * settings. This is run in a separate thread, and will only start if there
     * is no associator already running. The associator output is sent to the
     * results history panel.
     */
    protected void startRemoteAssociator() {

        /* =============== Weka4WS code removed ============== */

        // m_StartBut.setEnabled(false);
        m_StopBut.setEnabled(true);
        /* ============== Weka4WS modifications ============== */
        Thread runThread = new Thread() {
            /* =================================================== */
            public void run() {
                // Copy the current state of things
                m_Log.statusMessage("Setting up...");
                Instances inst = new Instances(m_Instances);

                Associator associator = (Associator) m_AssociatorEditor
                        .getValue();
                StringBuffer outBuff = new StringBuffer();
                String name = (new SimpleDateFormat("HH:mm:ss - "))
                        .format(new Date());
                String cname = associator.getClass().getName();
                if (cname.startsWith("weka.associations.")) {
                    name += cname.substring("weka.associations.".length());
                } else {
                    name += cname;
                }
                try {

                    /* =============== Weka4WS code removed ============== */
                    if (m_Log instanceof TaskLogger) {
                        ((TaskLogger) m_Log).taskStarted();
                    }
                    outBuff.append("=== Run information (" + remoteHost
                            + ") ===\n\n");
                    outBuff.append("Scheme:       " + cname);

                    /* ============== Weka4WS modifications ============== */
                    String optionsString = "";
                    if (associator instanceof OptionHandler) {
                        String[] optionsArray = ((OptionHandler) associator)
                                .getOptions();
                        optionsString = Utils.joinOptions(optionsArray);
                        outBuff.append(" " + optionsString);
                    }

                    initParameters(m_HostsCombo.getSelectedIndex());

                    ConnectionParameters connectionParameters = new ConnectionParameters(
                            remoteHost, remoteServicePort, remoteGridFTPPort,
                            remoteLogging, m_Log);
                    AssociationRulesParameters assRulesParameters = new AssociationRulesParameters();

                    RemoteAssociator remAssociator = new RemoteAssociator(
                            connectionParameters, assRulesParameters);

                    remAssociator.setAlgorithm(cname, optionsString);
                    /* =================================================== */

                    outBuff.append("\n");
                    outBuff.append("Relation:     " + inst.relationName()
                            + '\n');
                    outBuff.append("Instances:    " + inst.numInstances()
                            + '\n');
                    outBuff.append("Attributes:   " + inst.numAttributes()
                            + '\n');
                    if (inst.numAttributes() < 100) {
                        for (int i = 0; i < inst.numAttributes(); i++) {
                            outBuff.append("              "
                                    + inst.attribute(i).name() + '\n');
                        }
                    } else {
                        outBuff
                                .append("              [list of attributes omitted]\n");
                    }
                    /* ============== Weka4WS modifications ============== */
                    m_History.addResult(name, outBuff, remAssociator, this);
                    /* =================================================== */
                    m_History.setSingle(name);

                    /* ============== Weka4WS modifications ============== */
                    // Build the model and output it.
                    m_Log.statusMessage("Building model on " + remoteHost);

                    // for timing
                    long trainTimeStart = 0, trainTimeElapsed = 0;

                    trainTimeStart = System.currentTimeMillis();

                    saveInstancesToFile();

                    remAssociator.setDataset(m_FilePath);

                    Result result = remAssociator.getResult();

                    if (result == null) {
                        throw new TaskStoppedException();
                    }

                    trainTimeElapsed = System.currentTimeMillis()
                            - trainTimeStart;

                    ModelResult modelResult = result.getModel();

                    associator = (Associator) WUtil.getObjFromStr(modelResult
                            .getModel());

                    /* =================================================== */

                    outBuff
                            .append("=== Associator model (full training set) ===\n\n");
                    outBuff.append(associator.toString() + '\n');

                    /* ============== Weka4WS modifications ============== */
                    outBuff.append("\nTime taken to build model: "
                            + Utils
                                    .doubleToString(trainTimeElapsed / 1000.0,
                                            2) + " seconds\n\n");
                    /* =================================================== */

                    m_History.updateResult(name);
                    m_Log.logMessage("Finished " + cname);
                    m_Log.statusMessage("OK");
                }
                /* ============== Weka4WS modifications ============== */
                catch (TaskStoppedException ex) {
                    m_Log.logMessage("Associator interrupted");
                    m_Log.statusMessage("Associator interrupted");
                    outBuff.append("=== Associator interrupted ===");
                    m_History.updateResult(name);
                } catch (Exception ex) {
                    m_Log.logMessage(ex.getMessage());
                    WUtil
                            .showDialog(AssociationsPanel.this,
                                    ex.getMessage(), "Error on " + remoteHost,
                                    JOptionPane.ERROR_MESSAGE);
                    m_Log
                            .statusMessage("Problem building association rules (see log)");
                    outBuff
                            .append("=== Associator interrupted with errors ===");
                    m_History.updateResult(name);
                    /* =================================================== */
                } finally {
                    if (isInterrupted()) {
                        m_Log.logMessage("Interrupted " + cname);
                        m_Log.statusMessage("See error log");
                    }

                    /* =============== Weka4WS code removed ============== */

                    /* ============== Weka4WS modifications ============== */
                    if (m_History.isSelected(this)) {
                        m_StopBut.setEnabled(false);
                    }
                    /* =================================================== */
                    if (m_Log instanceof TaskLogger) {
                        ((TaskLogger) m_Log).taskFinished();
                    }
                }
            }
        };
        runThread.setPriority(Thread.MIN_PRIORITY);
        runThread.start();
    }

    /* =================================================== */

    /** Lets the user configure the associator */
    protected GenericObjectEditor m_AssociatorEditor = new GenericObjectEditor();

    /** The panel showing the current associator selection */
    protected PropertyPanel m_CEPanel = new PropertyPanel(m_AssociatorEditor);

    /** The output area for associations */
    protected JTextArea m_OutText = new JTextArea(20, 40);

    /* ============== Weka4WS modifications ============== */
    // /** The destination for log/status messages */
    // protected Logger m_Log = new SysErrLog();
    /* =================================================== */

    /** The buffer saving object for saving output */
    protected SaveBuffer m_SaveOut = new SaveBuffer(m_Log, this);

    /** A panel controlling results viewing */
    protected ResultHistoryPanel m_History = new ResultHistoryPanel(m_OutText);

    /* ============== Weka4WS modifications ============== */
    // /** Click to start running the associator */
    // protected JButton m_StartBut = new JButton("Start");
    // /** Click to stop a running associator */
    // protected JButton m_StopBut = new JButton("Stop");
    /* =================================================== */

    /* =============== Weka4WS code removed ============== */

    /** The user-supplied test set (if any) */
    protected Instances m_TestInstances;

    /** A thread that associator runs in */
    protected Thread m_RunThread;

    /* Register the property editors we need */
    static {
        GenericObjectEditor.registerEditors();
    }

    /**
     * Creates the associator panel
     */
    public AssociationsPanel() {

        /* ============== Weka4WS modifications ============== */
        super();
        /* =================================================== */

        // Connect / configure the components
        m_OutText.setEditable(false);
        m_OutText.setFont(new Font("Monospaced", Font.PLAIN, 12));
        m_OutText.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        m_OutText.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if ((e.getModifiers() & InputEvent.BUTTON1_MASK) != InputEvent.BUTTON1_MASK) {
                    m_OutText.selectAll();
                }
            }
        });
        m_History.setBorder(BorderFactory
                .createTitledBorder("Result list (right-click for options)"));
        m_History.setHandleRightClicks(false);
        // see if we can popup a menu for the selected result
        m_History.getList().addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                if (((e.getModifiers() & InputEvent.BUTTON1_MASK) != InputEvent.BUTTON1_MASK)
                        || e.isAltDown()) {
                    int index = m_History.getList().locationToIndex(
                            e.getPoint());
                    if (index != -1) {
                        String name = m_History.getNameAtIndex(index);
                        historyRightClickPopup(name, e.getX(), e.getY());
                    } else {
                        historyRightClickPopup(null, e.getX(), e.getY());
                    }
                }
                /* ============== Weka4WS modifications ============== */
                else if ((e.getModifiers() & InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK) {

                    int index = m_History.getList().locationToIndex(
                            e.getPoint());
                    if (m_History.isRunning(index)) {
                        m_Log.statusMessage("Associator running...");
                        m_StopBut.setEnabled(true);
                    } else {
                        m_Log.statusMessage("OK");
                        m_StopBut.setEnabled(false);
                    }

                }
                /* =================================================== */
            }
        });

        m_AssociatorEditor.setClassType(Associator.class);
        m_AssociatorEditor.setValue(new weka.associations.Apriori());
        m_AssociatorEditor
                .addPropertyChangeListener(new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        repaint();
                    }
                });

        m_StartBut.setToolTipText("Starts the associator");
        m_StopBut.setToolTipText("Stops the associator");

        m_StartBut.setEnabled(false);
        m_StopBut.setEnabled(false);
        m_StartBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                /* ============== Weka4WS modifications ============== */
                ButtonDisabler disabler = new ButtonDisabler(m_StartBut, 1000);
                disabler.start();
                String host = Hosts.getHostAt(m_HostsCombo.getSelectedIndex());
                if (host.equals(Hosts.LOCAL)) {
                    startAssociator();
                } else {
                    startRemoteAssociator();
                }
                /* =================================================== */
            }
        });
        m_StopBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                stopAssociator();
                m_StopBut.setEnabled(false);
                m_StartBut.setEnabled(true);
            }
        });

        // Layout the GUI
        JPanel p1 = new JPanel();
        p1.setBorder(BorderFactory.createCompoundBorder(BorderFactory
                .createTitledBorder("Associator"), BorderFactory
                .createEmptyBorder(0, 5, 5, 5)));
        p1.setLayout(new BorderLayout());
        p1.add(m_CEPanel, BorderLayout.NORTH);

        /* ============== Weka4WS modifications ============== */
        JPanel ssButs = new JPanel();
        ssButs.setLayout(new BoxLayout(ssButs, BoxLayout.X_AXIS));
        ssButs.add(m_StartBut);
        ssButs.add(Box.createGlue());
        ssButs.add(m_StopBut);
        ssButs.add(Box.createGlue());
        ssButs.add(m_ReloadBut);

        JPanel buttsAndCombo = new JPanel();
        buttsAndCombo.setBorder(new TitledBorder("Control panel"));
        GridLayout gl = new GridLayout(2, 1);
        gl.setVgap(5);
        buttsAndCombo.setLayout(gl);

        buttsAndCombo.add(ssButs);
        buttsAndCombo.add(m_HostsCombo);

        JPanel buttons = new JPanel();
        GridBagLayout gbL1 = new GridBagLayout();
        buttons.setLayout(gbL1);

        GridBagConstraints gbC1 = new GridBagConstraints();
        gbC1.anchor = GridBagConstraints.CENTER;
        gbC1.fill = GridBagConstraints.HORIZONTAL;
        gbC1.gridy = 0;
        gbC1.gridheight = 1;
        gbC1.weightx = 100;
        gbL1.setConstraints(buttsAndCombo, gbC1);
        buttons.add(buttsAndCombo);

        /* =================================================== */

        JPanel p3 = new JPanel();
        p3.setBorder(BorderFactory.createTitledBorder("Associator output"));
        p3.setLayout(new BorderLayout());
        final JScrollPane js = new JScrollPane(m_OutText);
        p3.add(js, BorderLayout.CENTER);
        js.getViewport().addChangeListener(new ChangeListener() {
            private int lastHeight;

            public void stateChanged(ChangeEvent e) {
                JViewport vp = (JViewport) e.getSource();
                int h = vp.getViewSize().height;
                if (h != lastHeight) { // i.e. an addition not just a user
                    // scrolling
                    lastHeight = h;
                    int x = h - vp.getExtentSize().height;
                    vp.setViewPosition(new Point(0, x));
                }
            }
        });

        GridBagLayout gbL = new GridBagLayout();
        GridBagConstraints gbC = new GridBagConstraints();
        JPanel mondo = new JPanel();
        gbL = new GridBagLayout();
        mondo.setLayout(gbL);
        gbC = new GridBagConstraints();
        gbC.anchor = GridBagConstraints.NORTH;
        gbC.fill = GridBagConstraints.HORIZONTAL;
        gbC.gridy = 1;
        gbC.gridx = 0;
        gbL.setConstraints(buttons, gbC);
        mondo.add(buttons);
        gbC = new GridBagConstraints();
        gbC.fill = GridBagConstraints.BOTH;
        gbC.gridy = 2;
        gbC.gridx = 0;
        gbC.weightx = 0;
        gbL.setConstraints(m_History, gbC);
        mondo.add(m_History);
        gbC = new GridBagConstraints();
        gbC.fill = GridBagConstraints.BOTH;
        gbC.gridy = 0;
        gbC.gridx = 1;
        gbC.gridheight = 3;
        gbC.weightx = 100;
        gbC.weighty = 100;
        gbL.setConstraints(p3, gbC);
        mondo.add(p3);

        setLayout(new BorderLayout());
        add(p1, BorderLayout.NORTH);
        add(mondo, BorderLayout.CENTER);

    }

    /**
     * Sets the Logger to receive informational messages
     * 
     * @param newLog
     *            the Logger that will now get info messages
     */
    public void setLog(Logger newLog) {

        m_Log = newLog;
    }

    /**
     * Tells the panel to use a new set of instances.
     * 
     * @param inst
     *            a set of Instances
     */
    /* ============== Weka4WS modifications ============== */
    public void setInstances(Instances inst, String fileURL, String filePath) {

        if (filePath == null) {
            m_FilePath = fileURL;
        } else {
            m_FilePath = filePath;
        }

        m_Instances = inst;
        String[] attribNames = new String[m_Instances.numAttributes()];
        for (int i = 0; i < attribNames.length; i++) {
            String type = "";
            switch (m_Instances.attribute(i).type()) {
            case Attribute.NOMINAL:
                type = "(Nom) ";
                break;
            case Attribute.NUMERIC:
                type = "(Num) ";
                break;
            case Attribute.STRING:
                type = "(Str) ";
                break;
            case Attribute.DATE:
                type = "(Dat) ";
                break;
            default:
                type = "(???) ";
            }
            attribNames[i] = type + m_Instances.attribute(i).name();
        }

        try {
            WUtil.loadThread.join();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        m_HostsCombo.setModel(Hosts.CBModel);
        m_HostsCombo.setSelectedIndex(0);

        m_StartBut.setEnabled(m_RunThread == null);
        m_StopBut.setEnabled(m_RunThread != null);
    }

    /* =================================================== */

    /**
     * Starts running the currently configured associator with the current
     * settings. This is run in a separate thread, and will only start if there
     * is no associator already running. The associator output is sent to the
     * results history panel.
     */
    protected void startAssociator() {

        if (m_RunThread == null) {

            /* =============== Weka4WS code removed ============== */

            m_StopBut.setEnabled(true);
            m_RunThread = new Thread() {
                public void run() {
                    // Copy the current state of things
                    m_Log.statusMessage("Setting up...");
                    Instances inst = new Instances(m_Instances);

                    Associator associator = (Associator) m_AssociatorEditor
                            .getValue();
                    StringBuffer outBuff = new StringBuffer();
                    String name = (new SimpleDateFormat("HH:mm:ss - "))
                            .format(new Date());
                    String cname = associator.getClass().getName();
                    if (cname.startsWith("weka.associations.")) {
                        name += cname.substring("weka.associations.".length());
                    } else {
                        name += cname;
                    }
                    try {

                        // Output some header information
                        m_Log.logMessage("Started " + cname);
                        if (m_Log instanceof TaskLogger) {
                            ((TaskLogger) m_Log).taskStarted();
                        }
                        outBuff
                                .append("=== Run information (localhost) ===\n\n");
                        outBuff.append("Scheme:       " + cname);
                        if (associator instanceof OptionHandler) {
                            String[] o = ((OptionHandler) associator)
                                    .getOptions();
                            outBuff.append(" " + Utils.joinOptions(o));
                        }
                        outBuff.append("\n");
                        outBuff.append("Relation:     " + inst.relationName()
                                + '\n');
                        outBuff.append("Instances:    " + inst.numInstances()
                                + '\n');
                        outBuff.append("Attributes:   " + inst.numAttributes()
                                + '\n');
                        if (inst.numAttributes() < 100) {
                            for (int i = 0; i < inst.numAttributes(); i++) {
                                outBuff.append("              "
                                        + inst.attribute(i).name() + '\n');
                            }
                        } else {
                            outBuff
                                    .append("              [list of attributes omitted]\n");
                        }
                        /* ============== Weka4WS modifications ============== */
                        m_History.addResult(name, outBuff, null, this);
                        /* =================================================== */
                        m_History.setSingle(name);

                        // Build the model and output it.
                        m_Log
                                .statusMessage("Building model on training data...");
                        associator.buildAssociations(inst);
                        outBuff
                                .append("=== Associator model (full training set) ===\n\n");
                        outBuff.append(associator.toString() + '\n');
                        m_History.updateResult(name);
                        m_Log.logMessage("Finished " + cname);
                        m_Log.statusMessage("OK");
                    } catch (Exception ex) {
                        m_Log.logMessage(ex.getMessage());
                        m_Log.statusMessage("See error log");
                    } finally {
                        if (isInterrupted()) {
                            m_Log.logMessage("Interrupted " + cname);
                            m_Log.statusMessage("See error log");
                        }
                        m_RunThread = null;
                        /* ============== Weka4WS modifications ============== */
                        if (m_History.isSelected(this)) {
                            m_StopBut.setEnabled(false);
                        }
                        /* =================================================== */
                        if (m_Log instanceof TaskLogger) {
                            ((TaskLogger) m_Log).taskFinished();
                        }
                    }
                }
            };
            m_RunThread.setPriority(Thread.MIN_PRIORITY);
            m_RunThread.start();
        }
    }

    /* ============== Weka4WS modifications ============== */
    /**
     * Stops the selected running classifier (if any).
     */
    protected void stopAssociator() {

        RemoteTask task = m_History.getSelectedRemoteTask();

        if (task == null) {
            if (m_RunThread != null) {
                m_RunThread.interrupt();
                // This is deprecated (and theoretically the interrupt should
                // do).
                m_RunThread.stop();
            }
        } else {
            try {
                task.stop();
            } catch (RemoteException e) {
                e.printStackTrace();
                m_Log.logMessage(e.getMessage());

                WUtil.showDialog(AssociationsPanel.this, e.getMessage(),
                        "Error stopping the associator",
                        JOptionPane.ERROR_MESSAGE);

                m_Log.statusMessage("Problem stopping associator");
            }
            m_History.setSelectedRemoteTaskToNull();
        }
    }

    /* =================================================== */

    /**
     * Save the currently selected associator output to a file.
     * 
     * @param name
     *            the name of the buffer to save
     */
    protected void saveBuffer(String name) {
        StringBuffer sb = m_History.getNamedBuffer(name);
        if (sb != null) {
            if (m_SaveOut.save(sb)) {
                m_Log.logMessage("Save successful.");
            }
        }
    }

    /**
     * Handles constructing a popup menu with visualization options.
     * 
     * @param name
     *            the name of the result history list entry clicked on by the
     *            user
     * @param x
     *            the x coordinate for popping up the menu
     * @param y
     *            the y coordinate for popping up the menu
     */
    protected void historyRightClickPopup(String name, int x, int y) {
        final String selectedName = name;
        JPopupMenu resultListMenu = new JPopupMenu();

        JMenuItem visMainBuffer = new JMenuItem("View in main window");
        if (selectedName != null) {
            visMainBuffer.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    m_History.setSingle(selectedName);
                }
            });
        } else {
            visMainBuffer.setEnabled(false);
        }
        resultListMenu.add(visMainBuffer);

        JMenuItem visSepBuffer = new JMenuItem("View in separate window");
        if (selectedName != null) {
            visSepBuffer.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    m_History.openFrame(selectedName);
                }
            });
        } else {
            visSepBuffer.setEnabled(false);
        }
        resultListMenu.add(visSepBuffer);

        JMenuItem saveOutput = new JMenuItem("Save result buffer");
        if (selectedName != null) {
            saveOutput.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    saveBuffer(selectedName);
                }
            });
        } else {
            saveOutput.setEnabled(false);
        }
        resultListMenu.add(saveOutput);

        resultListMenu.show(m_History.getList(), x, y);
    }

    /**
     * Tests out the Associator panel from the command line.
     * 
     * @param args
     *            may optionally contain the name of a dataset to load.
     */
    public static void main(String[] args) {

        try {
            final javax.swing.JFrame jf = new javax.swing.JFrame(
                    "Weka Explorer: Associator");
            jf.getContentPane().setLayout(new BorderLayout());
            final AssociationsPanel sp = new AssociationsPanel();
            jf.getContentPane().add(sp, BorderLayout.CENTER);
            weka.gui.LogPanel lp = new weka.gui.LogPanel();
            sp.setLog(lp);
            jf.getContentPane().add(lp, BorderLayout.SOUTH);
            jf.addWindowListener(new java.awt.event.WindowAdapter() {
                public void windowClosing(java.awt.event.WindowEvent e) {
                    jf.dispose();
                    System.exit(0);
                }
            });
            jf.pack();
            jf.setVisible(true);
            if (args.length == 1) {
                System.err.println("Loading instances from " + args[0]);
                java.io.Reader r = new java.io.BufferedReader(
                        new java.io.FileReader(args[0]));
                Instances i = new Instances(r);
                /* ============== Weka4WS modifications ============== */
                sp.setInstances(i, null, null);
                /* =================================================== */
            }
        } catch (Exception ex) {
            ex.printStackTrace();
            System.err.println(ex.getMessage());
        }
    }
}
