/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  ExplorerPanel.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.explorer;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;

import weka.core.Instances;
import weka.gui.Logger;
import weka.gui.SysErrLog;
import weka4ws.core.Hosts;
import weka4ws.core.NoHostsFoundException;
import weka4ws.core.WUtil;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public abstract class ExplorerPanel extends JPanel {

    private static final long serialVersionUID = 1L;

    // Lets the user select the remote grid node
    protected JComboBox m_HostsCombo = new JComboBox();

    protected JButton m_ReloadBut = new JButton("Reload hosts");

    protected String m_FilePath;

    protected String remoteHost = "";

    protected int remoteServicePort;

    protected int remoteGridFTPPort;

    protected int remoteLogging;

    protected static boolean preprocessed = false;

    protected static boolean loadedFromDB = false;

    /** The main set of instances we're playing with */
    protected Instances m_Instances;

    /** The destination for log/status messages */
    protected Logger m_Log = new SysErrLog();

    /** Click to start running the task */
    protected JButton m_StartBut = new JButton("Start");

    /** Click to stop a running task */
    protected JButton m_StopBut = new JButton("Stop");

    /** Stop the class combo from taking up to much space */
    protected Dimension COMBO_SIZE = new Dimension(150, m_StartBut
            .getPreferredSize().height);

    protected ExplorerPanel() {

        m_HostsCombo.setToolTipText("Selects the remote grid node");
        m_HostsCombo.setPreferredSize(COMBO_SIZE);
        m_HostsCombo.setMaximumSize(COMBO_SIZE);
        m_HostsCombo.setMinimumSize(COMBO_SIZE);

        m_ReloadBut
                .setToolTipText("Reloads the hosts file and check for their accessibility");
        m_ReloadBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (!Hosts.loading) {
                    Thread loadThread = new Thread() {
                        public void run() {
                            m_Log
                                    .statusMessage("Reloading hosts addresses, please wait...");
                            Hosts.loadAndCheck(true);
                            m_Log.statusMessage("Hosts addresses loaded.");
                        }
                    };
                    loadThread.start();
                }
            }
        });
    }

    protected void initParameters(int selectedHost) {
        String host = Hosts.getHostAt(selectedHost);
        if (host.equals(Hosts.AUTO)) {
            try {
                String[] parameters = new String[3];
                parameters = Hosts.getAutoHost();
                remoteHost = parameters[0];
                remoteServicePort = Integer.parseInt(parameters[1]);
                remoteGridFTPPort = Integer.parseInt(parameters[2]);
                remoteLogging = Integer.parseInt(parameters[3]);
            } catch (NoHostsFoundException e) {
                // TODO Mostrare un messaggio in finestra "No hosts found"
                // e fermare l'esecuzione del programma (cercare stop)
                e.printStackTrace();
            }
        } else {
            remoteHost = host;
            remoteServicePort = Integer
                    .parseInt(Hosts.servicePorts[selectedHost]);
            remoteGridFTPPort = Integer
                    .parseInt(Hosts.gridFTPPorts[selectedHost]);
            remoteLogging = Integer
                    .parseInt(Hosts.loggingOptions[selectedHost]);
        }

    }

    protected void saveInstancesToFile() {
        if (preprocessed && loadedFromDB) {
            m_Log
                    .statusMessage("Saving preprocessed database instances to file...");
            m_Log
                    .logMessage("Saving preprocessed database instances to file...");
            m_FilePath = WUtil.saveWorkingInstancesToFile("db", m_Instances);
            preprocessed = false;
            loadedFromDB = false;
        } else if (preprocessed) {
            m_Log.statusMessage("Saving preprocessed instances to file...");
            m_Log.logMessage("Saving preprocessed instances to file...");
            String fileName = m_FilePath.substring(m_FilePath
                    .lastIndexOf(File.separator) + 1, m_FilePath
                    .lastIndexOf("."));
            m_FilePath = WUtil
                    .saveWorkingInstancesToFile(fileName, m_Instances);
            preprocessed = false;
        } else if (loadedFromDB) {
            m_Log.statusMessage("Saving database instances to file...");
            m_Log.logMessage("Saving database instances to file...");
            m_FilePath = WUtil.saveWorkingInstancesToFile("db", m_Instances);
            loadedFromDB = false;
        }
    }

    // protected void enableStartBut() {
    // Thread enableThread = new Thread(){
    // public void run() {
    // try {
    //                    
    // wait(1100);
    // } catch (InterruptedException e) {
    // e.printStackTrace();
    // }
    // m_StartBut.setEnabled(true);
    // }
    // };
    // enableThread.start();
    // }

}
