/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    ClassAssignerBeanInfo.java
 *    Copyright (C) 2002 Mark Hall
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.beans.BeanDescriptor;
import java.beans.EventSetDescriptor;
import java.beans.PropertyDescriptor;
import java.beans.SimpleBeanInfo;

import weka.gui.beans.DataSource;
import weka.gui.beans.DataSourceListener;
import weka.gui.beans.InstanceListener;
import weka.gui.beans.TestSetListener;
import weka.gui.beans.TestSetProducer;
import weka.gui.beans.TrainingSetListener;
import weka.gui.beans.TrainingSetProducer;

/**
 * BeanInfo class for the class assigner bean
 *
 * @author <a href="mailto:mhall@cs.waikato.ac.nz">Mark Hall</a>
 * @author Marco Lackovic
 * @version $Revision: 1.1 $
 */
public class ClassAssignerBeanInfo extends SimpleBeanInfo {

    /**
     * Returns the event set descriptors
     *
     * @return an <code>EventSetDescriptor[]</code> value
     */
    public EventSetDescriptor[] getEventSetDescriptors() {
        try {
            EventSetDescriptor[] esds = {
                    new EventSetDescriptor(DataSource.class, "dataSet",
                            DataSourceListener.class, "acceptDataSet"),
                    new EventSetDescriptor(DataSource.class, "instance",
                            InstanceListener.class, "acceptInstance"),
                    new EventSetDescriptor(TrainingSetProducer.class,
                            "trainingSet", TrainingSetListener.class,
                            "acceptTrainingSet"),
                    new EventSetDescriptor(TestSetProducer.class, "testSet",
                            TestSetListener.class, "acceptTestSet") };
            return esds;
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return null;
    }

    /**
     * Returns the property descriptors
     *
     * @return a <code>PropertyDescriptor[]</code> value
     */
    public PropertyDescriptor[] getPropertyDescriptors() {
        try {
            PropertyDescriptor p1;
            p1 = new PropertyDescriptor("classColumn", ClassAssigner.class);
            PropertyDescriptor[] pds = { p1 };
            return pds;
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return null;
    }

    public BeanDescriptor getBeanDescriptor() {
        return new BeanDescriptor(
                weka4ws.gui.kflow.ClassAssigner.class,
                ClassAssignerCustomizer.class);
    }
}
