/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Clusterer.java
 *    Copyright (C) 2004 Stefan Mutter
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.awt.BorderLayout;
import java.beans.EventSetDescriptor;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import weka.clusterers.EM;
import weka.core.Instances;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Remove;
import weka.gui.Logger;
import weka.gui.beans.BatchClustererEvent;
import weka.gui.beans.BatchClustererListener;
import weka.gui.beans.BeanCommon;
import weka.gui.beans.BeanVisual;
import weka.gui.beans.DataSetEvent;
import weka.gui.beans.EventConstraints;
import weka.gui.beans.GraphEvent;
import weka.gui.beans.GraphListener;
import weka.gui.beans.TestSetEvent;
import weka.gui.beans.TestSetListener;
import weka.gui.beans.TextEvent;
import weka.gui.beans.TextListener;
import weka.gui.beans.TrainingSetListener;
import weka.gui.beans.UserRequestAcceptor;
import weka.gui.beans.Visible;
import weka.gui.beans.WekaWrapper;
import weka4ws.core.ClusteringParameters;
import weka4ws.core.ConnectionParameters;
import weka4ws.core.Hosts;
import weka4ws.core.RemoteClusterer;
import weka4ws.core.TaskStoppedException;
import weka4ws.core.WUtil;
import weka4ws.stubs.service.ArrayOfInt;
import weka4ws.stubs.service.Result;

/**
 * Bean that wraps around weka.clusterers
 * 
 * @author <a href="mailto:mutter@cs.waikato.ac.nz">Stefan Mutter</a>
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 * @version $Revision: 1.1.2.2 $
 * @see JPanel
 * @see BeanCommon
 * @see Visible
 * @see WekaWrapper
 * @see Serializable
 * @see UserRequestAcceptor
 * @see TrainingSetListener
 * @see TestSetListener
 */

/* ============== Weka4WS modifications ============== */
public class Clusterer extends ComputingNode implements BeanCommon, Visible,
        WekaWrapper, EventConstraints, UserRequestAcceptor,
        TrainingSetListener, TestSetListener {

    private static final long serialVersionUID = 1L;

    protected RemoteClusterer remClusterer = null;

    /**
     * Builds the clusters on a remote host
     */
    private void buildRemoteClusterer() throws Exception {

        ConnectionParameters connectionParameters = new ConnectionParameters(
                remoteHost, remoteServicePort, remoteGridFTPPort,
                remoteLogging, m_log);

        String optionsString = "";
        if (m_Clusterer instanceof OptionHandler) {
            String[] optionsArray = ((OptionHandler) m_Clusterer).getOptions();
            optionsString = Utils.joinOptions(optionsArray);
        }

        ClusteringParameters clusteringParameters = new ClusteringParameters();

        clusteringParameters.setClassIndex(m_trainingSet.classIndex());

        // TODO m_ClassesToClustersBut.isSelected

        // TODO ignore attributes

        int[] ignoredAtts = null;
        clusteringParameters.setSelectedAttributes(new ArrayOfInt(ignoredAtts));

        RemoteClusterer remClusterer = new RemoteClusterer(
                connectionParameters, clusteringParameters);
        remClusterer.setTestOptions(testOptions);
        remClusterer.setAlgorithm(cname, optionsString);
        remClusterer.setDataset(dataSetFilePath);

        Result result = remClusterer.getResult();
        if (result == null) {
            throw new TaskStoppedException();
        } else {
            modelResult = result.getModel();
            evalResult = result.getEvaluation();
            visResult = result.getVisualization();
        }

    }

    private void batchListenersCheck(final weka.gui.beans.TrainingSetEvent e) {
        if (m_batchClustererListeners.size() > 0) {
            if (hostSelectedIndex == Hosts.LOCAL_INDEX) {
                BatchClustererEvent ce = new BatchClustererEvent(this,
                        m_Clusterer,
                        new DataSetEvent(this, e.getTrainingSet()), e
                                .getSetNumber(), e.getMaxSetNumber(), 1);
                notifyBatchClustererListeners(ce);
            } else {
                BatchClustererEvent ce = new BatchClustererEvent(this,
                        m_Clusterer, new DataSetEvent4WS(this, e
                                .getTrainingSet(), ""), e.getSetNumber(), e
                                .getMaxSetNumber(), 1);
                String modelString = modelResult.getModel();
                weka.clusterers.Clusterer clusterer = (weka.clusterers.Clusterer) WUtil
                        .getObjFromStr(modelString);
                setClusterer(clusterer);
                notifyBatchClustererListeners(ce);
            }
        }

    }

    private void graphListenersCheck(final weka.gui.beans.TrainingSetEvent e)
            throws Exception {
        if (m_Clusterer instanceof weka.core.Drawable
                && m_graphListeners.size() > 0) {
            if (hostSelectedIndex != Hosts.LOCAL_INDEX) {
                String modelString = modelResult.getModel();
                weka.clusterers.Clusterer clusterer = (weka.clusterers.Clusterer) WUtil
                        .getObjFromStr(modelString);
                setClusterer(clusterer);
            }
            String grphString = ((weka.core.Drawable) m_Clusterer).graph();
            int grphType = ((weka.core.Drawable) m_Clusterer).graphType();
            String grphTitle = m_Clusterer.getClass().getName();
            grphTitle = grphTitle.substring(grphTitle.lastIndexOf('.') + 1,
                    grphTitle.length());
            grphTitle = "Set " + e.getSetNumber() + " ("
                    + e.getTrainingSet().relationName() + ") " + grphTitle;

            GraphEvent ge = new GraphEvent(Clusterer.this, grphString,
                    grphTitle, grphType);
            notifyGraphListeners(ge);
        }
    }

    private void textListenersCheck(final weka.gui.beans.TrainingSetEvent e) {
        if (m_textListeners.size() > 0) {
            if (hostSelectedIndex != Hosts.LOCAL_INDEX) {
                String modelString = modelResult.getModel();
                weka.clusterers.Clusterer clusterer = (weka.clusterers.Clusterer) WUtil
                        .getObjFromStr(modelString);
                setClusterer(clusterer);
            }
            String modelString = m_Clusterer.toString();
            String titleString = m_Clusterer.getClass().getName();

            titleString = titleString.substring(
                    titleString.lastIndexOf('.') + 1, titleString.length());
            modelString = "=== Clusterer model ===\n\n"
                    + "Scheme:   "
                    + titleString
                    + "\n"
                    + "Relation: "
                    + m_trainingSet.relationName()
                    + ((e.getMaxSetNumber() > 1) ? "\nTraining Fold: "
                            + e.getSetNumber() : "") + "\n\n" + modelString;
            titleString = "Model: " + titleString;

            TextEvent nt = new TextEvent(Clusterer.this, modelString,
                    titleString);
            notifyTextListeners(nt);
        }

    }

    /**
     * Notify all batch clusterer listeners of the last batch clusterer event
     * 
     * @param ce
     *            a <code>BatchClustererEvent</code> value
     */
    private void lastNotifyBatchClustererListeners(BatchClustererEvent ce) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_batchClustererListeners.clone();
        }
        if (l.size() > 0) {
            if (hostSelectedIndex == Hosts.LOCAL_INDEX
                    || ce.getTestSet().isStructureOnly()) {
                for (int i = 0; i < l.size(); i++) {
                    // if (i == (l.size() - 1)) {
                    // ((ClustererPerformanceEvaluator) l.elementAt(i))
                    // .setTrainTotalTimeStart(getTrainTotalTimeStart());
                    // }
                    ((BatchClustererListener) l.elementAt(i))
                            .acceptClusterer(ce);
                }
            } else {
                for (int i = 0; i < l.size(); i++) {
                    // if (i == (l.size() - 1)) {
                    // ((ClustererPerformanceEvaluator) l.elementAt(i))
                    // .setTrainTotalTimeStart(getTrainTotalTimeStart());
                    // }
                    ((BatchRemoteClustererListener) l.elementAt(i))
                            .acceptRemoteClusterer(ce, remoteHost);
                }
            }
        }
    }

    /* =================================================== */

    protected BeanVisual m_visual = new BeanVisual("Clusterer",
            BeanVisual.ICON_PATH + "EM.gif", BeanVisual.ICON_PATH
                    + "EM_animated.gif");

    private Thread m_buildThread = null;

    /**
     * Global info for the wrapped classifier (if it exists).
     */
    protected String m_globalInfo;

    /**
     * Objects talking to us
     */
    private Hashtable m_listenees = new Hashtable();

    /**
     * Objects listening for batch clusterer events
     */
    private Vector m_batchClustererListeners = new Vector();

    /**
     * Objects listening for graph events
     */
    private Vector m_graphListeners = new Vector();

    /**
     * Objects listening for text events
     */
    private Vector m_textListeners = new Vector();

    /**
     * Holds training instances for batch training.
     */
    private Instances m_trainingSet;

    private transient Instances m_testingSet;

    private weka.clusterers.Clusterer m_Clusterer = new EM();

    private Double m_dummy = new Double(0.0);

    /**
     * Global info (if it exists) for the wrapped classifier
     * 
     * @return the global info
     */
    public String globalInfo() {
        return m_globalInfo;
    }

    /**
     * Creates a new <code>Clusterer</code> instance.
     */
    public Clusterer() {
        setLayout(new BorderLayout());
        add(m_visual, BorderLayout.CENTER);
        setClusterer(m_Clusterer);
    }

    /**
     * Set the clusterer for this wrapper
     * 
     * @param c
     *            a <code>weka.clusterers.Clusterer</code> value
     */
    public void setClusterer(weka.clusterers.Clusterer c) {
        boolean loadImages = true;
        if (c.getClass().getName().compareTo(m_Clusterer.getClass().getName()) == 0) {
            loadImages = false;
        } else {
            // clusterer has changed so any batch training status is now
            // invalid
            m_trainingSet = null;
        }
        m_Clusterer = c;
        String clustererName = c.getClass().toString();
        clustererName = clustererName.substring(
                clustererName.lastIndexOf('.') + 1, clustererName.length());
        if (loadImages) {
            if (!m_visual.loadIcons(BeanVisual.ICON_PATH + clustererName
                    + ".gif", BeanVisual.ICON_PATH + clustererName
                    + "_animated.gif")) {
                useDefaultVisual();
            }
        }
        m_visual.setText(clustererName);

        // get global info
        m_globalInfo = KnowledgeFlowApp.getGlobalInfo(m_Clusterer);
    }

    /**
     * Returns true if this clusterer has an incoming connection that is a batch
     * set of instances
     * 
     * @return a <code>boolean</code> value
     */
    public boolean hasIncomingBatchInstances() {
        if (m_listenees.size() == 0) {
            return false;
        }
        if (m_listenees.containsKey("trainingSet")
                || m_listenees.containsKey("testSet")
                || m_listenees.containsKey("dataSet")) {
            return true;
        }
        return false;
    }

    /**
     * Get the clusterer currently set for this wrapper
     * 
     * @return a <code>weka.clusterers.Clusterer</code> value
     */
    public weka.clusterers.Clusterer getClusterer() {
        return m_Clusterer;
    }

    /**
     * Sets the algorithm (clusterer) for this bean
     * 
     * @param algorithm
     *            an <code>Object</code> value
     * @exception IllegalArgumentException
     *                if an error occurs
     */
    public void setWrappedAlgorithm(Object algorithm) {

        if (!(algorithm instanceof weka.clusterers.Clusterer)) {
            throw new IllegalArgumentException(algorithm.getClass()
                    + " : incorrect " + "type of algorithm (Clusterer)");
        }
        setClusterer((weka.clusterers.Clusterer) algorithm);
    }

    /**
     * Returns the wrapped clusterer
     * 
     * @return an <code>Object</code> value
     */
    public Object getWrappedAlgorithm() {
        return getClusterer();
    }

    /**
     * Accepts a training set and builds batch clusterer
     * 
     * @param e
     *            a <code>TrainingSetEvent</code> value
     */
    public void acceptTrainingSet(final weka.gui.beans.TrainingSetEvent e) {
        /* ============== Weka4WS modifications ============== */
        dataSetFilePath = ((TrainingSetEvent4WS) e).getFilePath();
        if (e.isStructureOnly()) {
            // no need to build a clusterer, instead just generate a dummy
            // BatchClustererEvent in order to pass on instance structure to
            // any listeners
            DataSetEvent4WS dse4ws = new DataSetEvent4WS(this, e
                    .getTrainingSet(), ((TrainingSetEvent4WS) e).getFilePath());
            BatchClustererEvent ce = new BatchClustererEvent(this, m_Clusterer,
                    dse4ws, e.getSetNumber(), e.getMaxSetNumber(), 1);

            notifyBatchClustererListeners(ce);
            return;
        }

        initParameters();

        if ((m_buildThread == null)
                && (hostSelectedIndex == Hosts.LOCAL_INDEX || e.getSetNumber() < 2)) {
            try {
                if (m_state == IDLE) {
                    synchronized (this) {
                        m_state = BUILDING_MODEL;
                    }
                    m_trainingSet = e.getTrainingSet();
                    testOptions = ((TrainingSetEvent4WS) e).getTestOptions();
                    testMode = testOptions.getTestMode();
                    final String oldText = m_visual.getText();
                    m_buildThread = new Thread() {
                        public void run() {
                            increaseCounter();
                            try {
                                if (m_trainingSet != null) {
                                    m_visual.setAnimated();
                                    if (m_log != null) {
                                        m_log
                                                .statusMessage("Clusterer : building clusters...");
                                    }

                                    buildClusterer();

                                    batchListenersCheck(e);
                                    graphListenersCheck(e);
                                    textListenersCheck(e);
                                    setModelBuilt(true);
                                    if (connectedTSProducer != null) {
                                        connectedTSProducer
                                                .notifyClassifierBuilt();
                                    }
                                }
                            } catch (TaskStoppedException ex) {
                                m_log.logMessage("Clustering interrupted");
                                m_log.statusMessage("Clustering interrupted");
                            } catch (Exception ex) {
                                if (ex == null || ex.getMessage() == null
                                        || ex.getMessage().length() != 0) {
                                    ex.printStackTrace();
                                    if (m_log != null) {
                                        m_log
                                                .statusMessage("There has been errors: see log for details");
                                    }
                                    final String message = ex.getMessage();
                                    Thread dialogThread = new Thread() {
                                        public synchronized void run() {
                                            WUtil.showDialog(Clusterer.this,
                                                    message, "Error on "
                                                            + remoteHost,
                                                    JOptionPane.ERROR_MESSAGE);
                                        }
                                    };
                                    dialogThread.start();

                                } else {
                                    System.err
                                            .println("Empty exception message");
                                }

                            } finally {
                                decreaseCounter();
                                lastOneCheck();
                                m_visual.setText(oldText);
                                m_visual.setStatic();
                                m_state = IDLE;
                                if (isInterrupted()) {
                                    // prevent any clusterer events from being
                                    // fired
                                    m_trainingSet = null;
                                    if (m_log != null) {
                                        // TODO Quando l'utente interrompe le
                                        // operazioni,
                                        // o chiude il programma, o il programma
                                        // crasha bisogna
                                        // avvertire il servizio di fermarsi
                                        m_log
                                                .logMessage("Build clusterer interrupted!");
                                        m_log
                                                .statusMessage("Computation has been interrupted by the user");
                                    }
                                } else {
                                    // save header
                                    m_trainingSet = new Instances(
                                            m_trainingSet, 0);
                                }
                                block(false);
                                m_buildThread = null;
                            }
                        }
                    };
                    m_buildThread.setPriority(Thread.MIN_PRIORITY);
                    m_buildThread.start();
                    // make sure the thread is still running before we block
                    // if (m_buildThread.isAlive()) {
                    // block(true);
                    // }
                    // m_buildThread = null;
                    // m_state = IDLE;
                    /* =================================================== */
                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
    }

    /* ============== Weka4WS modifications ============== */
    /**
     * Accepts a test set for a batch trained clusterer
     * 
     * @param e
     *            a <code>TestSetEvent</code> value
     */
    public void acceptTestSet(TestSetEvent e) {

        if (m_trainingSet != null) {
            if (hostSelectedIndex == Hosts.LOCAL_INDEX || e.getSetNumber() < 2) {
                try {
                    // TODO Questo controllo pu essere evitato in quanto
                    // effettuato gi da TestSetNotifierThread
                    if (!isModelBuilt()) {
                        block(true);
                    }
                    if (m_state == IDLE) {
                        synchronized (this) {
                            m_state = CLUSTERING;
                        }
                        m_testingSet = e.getTestSet();
                        if (m_trainingSet.equalHeaders(m_testingSet)) {
                            BatchClustererEvent ce = new BatchClustererEvent(
                                    this, m_Clusterer, new DataSetEvent4WS(
                                            this, e.getTestSet(), ""), e
                                            .getSetNumber(), e
                                            .getMaxSetNumber(), 0);
                            if (m_batchClustererListeners.size() > 0) {
                                if (isLastOne()) {
                                    lastNotifyBatchClustererListeners(ce);
                                } else {
                                    System.out
                                            .println("acceptTestSet: notifyBatchClustererListeners");
                                    notifyBatchClustererListeners(ce);
                                }
                            }
                        }
                        synchronized (this) {
                            m_state = IDLE;
                        }
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    /**
     * Builds the clusters locally
     */
    private void buildClusterer() throws Exception {
        // fully Qualified Classifier Class Name
        cname = m_Clusterer.getClass().getName();
        m_visual.setText("Building clusters on " + remoteHost);

        if (hostSelectedIndex == Hosts.LOCAL_INDEX) {
            String dataSetFileName = dataSetFilePath.substring(dataSetFilePath
                    .lastIndexOf('/') + 1);
            String message = "Data mining of dataset \"" + dataSetFileName
                    + "\"";
            m_log.statusMessage(message);
            logMessage(message);
            long trainTimeStart = System.currentTimeMillis();

            if (m_trainingSet.classIndex() < 0) {
                m_Clusterer.buildClusterer(m_trainingSet);
            } else { // class based evaluation if class attribute is set
                Remove removeClass = new Remove();
                removeClass.setAttributeIndices(""
                        + (m_trainingSet.classIndex() + 1));
                removeClass.setInvertSelection(false);
                removeClass.setInputFormat(m_trainingSet);
                Instances clusterTrain = Filter.useFilter(m_trainingSet,
                        removeClass);
                m_Clusterer.buildClusterer(clusterTrain);
            }

            long trainTimeElapsed = System.currentTimeMillis() - trainTimeStart;
            logMessage("Time taken to build "
                    + cname.substring(cname.lastIndexOf('.') + 1) + " model: "
                    + Utils.doubleToString(trainTimeElapsed / 1000.0, 2)
                    + " seconds");
        } else {
            buildRemoteClusterer();
        }
    }

    /* =================================================== */

    /**
     * Sets the visual appearance of this wrapper bean
     * 
     * @param newVisual
     *            a <code>BeanVisual</code> value
     */
    public void setVisual(BeanVisual newVisual) {
        m_visual = newVisual;
    }

    /**
     * Gets the visual appearance of this wrapper bean
     */
    public BeanVisual getVisual() {
        return m_visual;
    }

    /**
     * Use the default visual appearance for this bean
     */
    public void useDefaultVisual() {
        m_visual.loadIcons(BeanVisual.ICON_PATH + "DefaultClusterer.gif",
                BeanVisual.ICON_PATH + "DefaultClusterer_animated.gif");
    }

    /**
     * Add a batch clusterer listener
     * 
     * @param cl
     *            a <code>BatchClustererListener</code> value
     */
    public synchronized void addBatchClustererListener(BatchClustererListener cl) {
        m_batchClustererListeners.addElement(cl);
    }

    /**
     * Remove a batch clusterer listener
     * 
     * @param cl
     *            a <code>BatchClustererListener</code> value
     */
    public synchronized void removeBatchClustererListener(
            BatchClustererListener cl) {
        m_batchClustererListeners.remove(cl);
    }

    /**
     * Notify all batch clusterer listeners of a batch clusterer event
     * 
     * @param ce
     *            a <code>BatchClustererEvent</code> value
     */
    private void notifyBatchClustererListeners(BatchClustererEvent ce) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_batchClustererListeners.clone();
        }
        if (l.size() > 0) {
            /* ============== Weka4WS modifications ============== */
            if (hostSelectedIndex == Hosts.LOCAL_INDEX
                    || ce.getTestSet().isStructureOnly()) {
                for (int i = 0; i < l.size(); i++) {
                    ((BatchClustererListener) l.elementAt(i))
                            .acceptClusterer(ce);
                }
            } else {
                for (int i = 0; i < l.size(); i++) {
                    ((BatchRemoteClustererListener) l.elementAt(i))
                            .acceptRemoteClusterer(ce, remoteHost);
                }
            }
            /* =================================================== */
        }
    }

    /**
     * Add a graph listener
     * 
     * @param cl
     *            a <code>GraphListener</code> value
     */
    public synchronized void addGraphListener(GraphListener cl) {
        m_graphListeners.addElement(cl);
    }

    /**
     * Remove a graph listener
     * 
     * @param cl
     *            a <code>GraphListener</code> value
     */
    public synchronized void removeGraphListener(GraphListener cl) {
        m_graphListeners.remove(cl);
    }

    /**
     * Notify all graph listeners of a graph event
     * 
     * @param ge
     *            a <code>GraphEvent</code> value
     */
    private void notifyGraphListeners(GraphEvent ge) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_graphListeners.clone();
        }
        if (l.size() > 0) {
            for (int i = 0; i < l.size(); i++) {
                ((GraphListener) l.elementAt(i)).acceptGraph(ge);
            }
        }
    }

    /**
     * Add a text listener
     * 
     * @param cl
     *            a <code>TextListener</code> value
     */
    public synchronized void addTextListener(TextListener cl) {
        m_textListeners.addElement(cl);
    }

    /**
     * Remove a text listener
     * 
     * @param cl
     *            a <code>TextListener</code> value
     */
    public synchronized void removeTextListener(TextListener cl) {
        m_textListeners.remove(cl);
    }

    /**
     * Notify all text listeners of a text event
     * 
     * @param ge
     *            a <code>TextEvent</code> value
     */
    private void notifyTextListeners(TextEvent ge) {
        Vector l;
        synchronized (this) {
            l = (Vector) m_textListeners.clone();
        }
        if (l.size() > 0) {
            for (int i = 0; i < l.size(); i++) {
                ((TextListener) l.elementAt(i)).acceptText(ge);
            }
        }
    }

    /**
     * Returns true if, at this time, the object will accept a connection with
     * respect to the named event
     * 
     * @param eventName
     *            the event
     * @return true if the object will accept a connection
     */
    public boolean connectionAllowed(String eventName) {
        /*
         * if (eventName.compareTo("instance") == 0) { if (!(m_Clusterer
         * instanceof weka.classifiers.UpdateableClassifier)) { return false; } }
         */
        if (m_listenees.containsKey(eventName)) {
            return false;
        }
        return true;
    }

    /**
     * Returns true if, at this time, the object will accept a connection
     * according to the supplied EventSetDescriptor
     * 
     * @param esd
     *            the EventSetDescriptor
     * @return true if the object will accept a connection
     */
    public boolean connectionAllowed(EventSetDescriptor esd) {
        return connectionAllowed(esd.getName());
    }

    /**
     * Notify this object that it has been registered as a listener with a
     * source with respect to the named event
     * 
     * @param eventName
     *            the event
     * @param source
     *            the source with which this object has been registered as a
     *            listener
     */
    public synchronized void connectionNotification(String eventName,
            Object source) {

        if (connectionAllowed(eventName)) {
            m_listenees.put(eventName, source);
            /*
             * if (eventName.compareTo("instance") == 0) {
             * startIncrementalHandler(); }
             */
        }
    }

    /**
     * Notify this object that it has been deregistered as a listener with a
     * source with respect to the supplied event name
     * 
     * @param eventName
     *            the event
     * @param source
     *            the source with which this object has been registered as a
     *            listener
     */
    public synchronized void disconnectionNotification(String eventName,
            Object source) {
        m_listenees.remove(eventName);
    }

    /**
     * Function used to stop code that calls acceptTrainingSet. This is needed
     * as clusterer construction is performed inside a separate thread of
     * execution.
     * 
     * @param tf
     *            a <code>boolean</code> value
     */
    private synchronized void block(boolean tf) {

        if (tf) {
            try {
                // only block if thread is still doing something useful!
                if (m_buildThread.isAlive() && m_state != IDLE) {
                    wait();
                }
            } catch (InterruptedException ex) {
            }
        } else {
            notifyAll();
        }
    }

    /**
     * Stop any clusterer action
     */
    public void stop() {

        /* ============== Weka4WS modifications ============== */
        if (remClusterer != null) {
            try {
                remClusterer.stop();
            } catch (RemoteException e) {
                e.printStackTrace();
                m_log.logMessage(e.getMessage());

                WUtil.showDialog(Clusterer.this, e.getMessage(),
                        "Error stopping the clusterer",
                        JOptionPane.ERROR_MESSAGE);

                m_log.statusMessage("Problem stopping clusterer");
            }
        }
        /* =================================================== */

        // tell all listenees (upstream beans) to stop
        Enumeration en = m_listenees.keys();
        while (en.hasMoreElements()) {
            Object tempO = m_listenees.get(en.nextElement());
            if (tempO instanceof BeanCommon) {
                /* ============== Weka4WS modifications ============== */
                // System.err.println("Listener is BeanCommon");
                /* =================================================== */
                ((BeanCommon) tempO).stop();
            }
        }

        // stop the build thread
        if (m_buildThread != null) {
            m_buildThread.interrupt();
            m_buildThread.stop();
            m_buildThread = null;
            m_visual.setStatic();
        }
    }

    /**
     * Set a logger
     * 
     * @param logger
     *            a <code>Logger</code> value
     */
    public void setLog(Logger logger) {
        m_log = logger;
    }

    /**
     * Return an enumeration of requests that can be made by the user
     * 
     * @return an <code>Enumeration</code> value
     */
    public Enumeration enumerateRequests() {
        Vector newVector = new Vector(0);
        if (m_buildThread != null) {
            newVector.addElement("Stop");
        }
        return newVector.elements();
    }

    /**
     * Perform a particular request
     * 
     * @param request
     *            the request to perform
     * @exception IllegalArgumentException
     *                if an error occurs
     */
    public void performRequest(String request) {
        if (request.compareTo("Stop") == 0) {
            stop();
        } else {
            throw new IllegalArgumentException(request
                    + " not supported (Classifier)");
        }
    }

    /**
     * Returns true, if at the current time, the event described by the supplied
     * event descriptor could be generated.
     * 
     * @param esd
     *            an <code>EventSetDescriptor</code> value
     * @return a <code>boolean</code> value
     */
    public boolean eventGeneratable(EventSetDescriptor esd) {
        String eventName = esd.getName();
        return eventGeneratable(eventName);
    }

    /**
     * Returns true, if at the current time, the named event could be generated.
     * Assumes that the supplied event name is an event that could be generated
     * by this bean
     * 
     * @param eventName
     *            the name of the event in question
     * @return true if the named event could be generated at this point in time
     */
    public boolean eventGeneratable(String eventName) {
        if (eventName.compareTo("graph") == 0) {
            // can't generate a GraphEvent if clusterer is not drawable
            if (!(m_Clusterer instanceof weka.core.Drawable)) {
                return false;
            }
            // need to have a training set before the clusterer
            // can generate a graph!
            if (!m_listenees.containsKey("trainingSet")) {
                return false;
            }
            // Source needs to be able to generate a trainingSet
            // before we can generate a graph
            Object source = m_listenees.get("trainingSet");
            if (source instanceof EventConstraints) {
                if (!((EventConstraints) source)
                        .eventGeneratable("trainingSet")) {
                    return false;
                }
            }
        }

        if (eventName.compareTo("batchClusterer") == 0) {
            if (!m_listenees.containsKey("trainingSet")) {
                return false;
            }

            Object source = m_listenees.get("trainingSet");
            if (source != null && source instanceof EventConstraints) {
                if (!((EventConstraints) source)
                        .eventGeneratable("trainingSet")) {
                    return false;
                }
            }
        }

        if (eventName.compareTo("text") == 0) {
            if (!m_listenees.containsKey("trainingSet")) {
                return false;
            }
            Object source = m_listenees.get("trainingSet");
            if (source != null && source instanceof EventConstraints) {
                if (!((EventConstraints) source)
                        .eventGeneratable("trainingSet")) {
                    return false;
                }
            }
        }

        if (eventName.compareTo("batchClassifier") == 0)
            return false;
        if (eventName.compareTo("incrementalClassifier") == 0)
            return false;

        return true;
    }

}
