/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    ClustererCustomizer.java
 *    Copyright (C) 2004 Stefan Mutter
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.awt.BorderLayout;
import java.beans.Customizer;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import javax.swing.JPanel;
import weka.gui.GenericObjectEditor;
import weka.gui.PropertySheetPanel;

/**
 * GUI customizer for the Clusterer wrapper bean
 * @author  <a href="mailto:mutter@cs.waikato.ac.nz">Stefan Mutter</a>
 * @author  Marco Lackovic
 * @version  $Revision: 1.1.2.1 $
 */
public class ClustererCustomizer extends JPanel implements Customizer {

    /* ============== Weka4WS modifications ============== */
    private static final long serialVersionUID = 1L;
    /* =================================================== */

    static {
        GenericObjectEditor.registerEditors();
    }

    private PropertyChangeSupport m_pcSupport = new PropertyChangeSupport(this);

    /* ============== Weka4WS modifications ============== */
    private weka4ws.gui.kflow.Clusterer m_dsClusterer;

    /* =================================================== */

    private PropertySheetPanel m_ClustererEditor = new PropertySheetPanel();

    public ClustererCustomizer() {

        setLayout(new BorderLayout());
        add(m_ClustererEditor, BorderLayout.CENTER);
    }

    /**
     * Set the Clusterer object to be edited
     * 
     * @param object
     *            an <code>Object</code> value
     */
    public void setObject(Object object) {
        /* ============== Weka4WS modifications ============== */
        m_dsClusterer = (weka4ws.gui.kflow.Clusterer) object;
        /* =================================================== */
        m_ClustererEditor.setTarget(m_dsClusterer.getClusterer());

    }

    /**
     * Add a property change listener
     * 
     * @param pcl
     *            a <code>PropertyChangeListener</code> value
     */
    public void addPropertyChangeListener(PropertyChangeListener pcl) {
        m_pcSupport.addPropertyChangeListener(pcl);
    }

    /**
     * Remove a property change listener
     * 
     * @param pcl
     *            a <code>PropertyChangeListener</code> value
     */
    public void removePropertyChangeListener(PropertyChangeListener pcl) {
        m_pcSupport.removePropertyChangeListener(pcl);
    }
}
