/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    KnowledgeFlow.java
 *    Copyright (C) 2005 Mark Hall
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.awt.Toolkit;

import weka.gui.beans.StartUpListener;
import weka4ws.core.WUtil;

/**
 * Startup class for the KnowledgeFlow. Displays a splash screen.
 * 
 * @author Mark Hall
 * @author Marco Lackovic
 * @version $Revision: 1.23.2.4 $
 */
public class KnowledgeFlow {

    /**
     * Static method that can be called from a running program to launch the
     * KnowledgeFlow
     */
    public static void startApp() {
        KnowledgeFlowApp.addStartupListener(new StartUpListener() {
            public void startUpComplete() {
                weka.gui.SplashWindow.disposeSplash();
            }
        });

        /* ============== Weka4WS modifications ============== */
        String splashImagePath = System.getProperty("user.dir")
                + "/img/splash.jpg";
        weka.gui.SplashWindow.splash(Toolkit.getDefaultToolkit().createImage(
                splashImagePath));
        /* =================================================== */

        Thread nt = new Thread() {
            public void run() {
                //                weka.gui.SplashWindow.invokeMethod(
                //                        "it.unical.deis.grid.weka4ws.kflow.KnowledgeFlowApp",
                //                        "createSingleton", new String[1]);
                KnowledgeFlowApp.createSingleton(new String[1]);
            }
        };
        nt.start();
    }

    /**
     * Shows the splash screen, launches the application and then disposes the
     * splash screen.
     * 
     * @param args
     *            the command line arguments
     */
    public static void main(String[] args) {

        /* ============== Weka4WS modifications ============== */
        WUtil.initialization();
        /* =================================================== */

        weka.gui.SplashWindow.splash(ClassLoader
                .getSystemResource("weka/gui/beans/icons/splash.jpg"));
        weka.gui.SplashWindow.invokeMain(
                "it.unical.deis.grid.weka4ws.kflow.KnowledgeFlowApp", args);
        weka.gui.SplashWindow.disposeSplash();
    }

}
