/*
 *  TestSetNotifierThread.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import weka.gui.beans.TestSetListener;

/**
 * Stops code that calls <code>notifyTestSetProduced</code>. This is needed as classifier
 * construction might not be completed yet when <code>notifyTestSetProduced</code> is invoked.
 * After its completion a <code>ComputingNode</code> will invoke <code>notifyClassifierBuilt</code>
 * on the instance of this class, which has previously registered itself by invoking the method
 * <code>registerForNotification</code>.
 */

public class TestSetNotifierThread extends Thread {

    private TestSetListener listener;

    private TestSetEvent4WS event;

    public TestSetNotifierThread(TestSetListener l, TestSetEvent4WS e) {
        listener = l;
        event = new TestSetEvent4WS(e.getSource(), e.getTestSet());
        event.setFilePath(e.getFilePath());
        event.setMaxSetNumber(e.getMaxSetNumber());
        event.setSetNumber(1);
        event.setTestOptions(e.getTestOptions());
    }

    /**
     * Registers itself for notification at the corresponding computing node
     * (a <code>Classifier</code> or a <code>Clusterer</code> 
     * 
     */
    public void run() {
        if (listener instanceof ComputingNode) {
            ComputingNode cn = ((ComputingNode) listener);
            cn.registerForNotification(this);
            while (!cn.isModelBuilt())
                block();
            cn.unregister(this);
            listener.acceptTestSet(event);
        } else
            listener.acceptTestSet(event);
    }

    /**
     * Notifies all the <code>TestSetNotifierThread</code>s that a
     * classifier has been built.
     * 
     */
    public synchronized void notifyClassifierBuilt() {
        notifyAll();
    }

    private synchronized void block() {
        try {
            wait();
        } catch (InterruptedException ex) {
        }
    }

}
