/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    TrainingSetMaker.java
 *    Copyright (C) 2002 Mark Hall
 *
 */

/*
 *  This code has been partially modified and extended as part of Weka4WS.
 *  Modifications and extensions are distinctly marked.
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.gui.kflow;

import java.io.Serializable;
import java.util.Vector;

import weka.gui.beans.AbstractTrainingSetProducer;
import weka.gui.beans.BeanVisual;
import weka.gui.beans.DataSetEvent;
import weka.gui.beans.DataSourceListener;
import weka.gui.beans.EventConstraints;
import weka.gui.beans.TrainingSetListener;
import weka4ws.core.RemoteTask;
import weka4ws.stubs.service.TestOptions;

/**
 * Bean that accepts a data sets and produces a training set
 * 
 * @author <a href="mailto:mhall@cs.waikato.ac.nz">Mark Hall</a>
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 * @version $Revision: 1.3 $
 */
public class TrainingSetMaker extends AbstractTrainingSetProducer implements
        DataSourceListener, EventConstraints, Serializable {

    /* ============== Weka4WS modifications ============== */

    private static final long serialVersionUID = 1L;

    /* =================================================== */

    public TrainingSetMaker() {
        m_visual.loadIcons(BeanVisual.ICON_PATH + "TrainingSetMaker.gif",
                BeanVisual.ICON_PATH + "TrainingSetMaker_animated.gif");
        m_visual.setText("TrainingSetMaker");
    }

    /**
     * Global info for this bean
     * 
     * @return a <code>String</code> value
     */
    public String globalInfo() {
        return "Designate an incoming data set as a training set.";
    }

    /* ============== Weka4WS modifications ============== */
    /**
     * Accept a data set
     * 
     * @param e
     *            a <code>DataSetEvent</code> value
     */
    public void acceptDataSet(DataSetEvent e) {

        // System.err.println("In accept data set as training set");
        TrainingSetEvent4WS tse = new TrainingSetEvent4WS(this, e.getDataSet());
        tse.setFilePath(((DataSetEvent4WS) e).getFilePath());
        
        tse.setSetNumber(1);
        tse.setMaxSetNumber(1);
        
        TestOptions testOptions = new TestOptions();
        testOptions.setTestMode(RemoteTask.TRAINING_SET);
        testOptions.setPercent(0);
        testOptions.setNumFolds(0);
        
        tse.setTestOptions(testOptions);
        
        notifyTrainingSetProduced(tse);
    }

    /**
     * Inform training set listeners that a training set is availabel
     * 
     * @param tse
     *            a <code>TrainingSetEvent</code> value
     */
    protected void notifyTrainingSetProduced(TrainingSetEvent4WS tse) {

        Vector l;
        synchronized (this) {
            l = (Vector) m_listeners.clone();
        }
        if (l.size() > 0) {

            if (tse.isStructureOnly()) {
                for (int i = 0; i < l.size(); i++) {
                    ((TrainingSetListener) l.elementAt(i))
                            .acceptTrainingSet(tse);
                }

            } else {
                for (int i = 0; i < l.size(); i++) {
                    TrainingSetNotifierThread notifyThread = new TrainingSetNotifierThread(
                            ((TrainingSetListener) l.elementAt(i)), tse);
                    notifyThread.start();
                }
            }
        }
    }

    /* =================================================== */

    /**
     * Stop any action
     */
    public void stop() {
        // do something
    }

    /**
     * Returns true, if at the current time, the named event could be generated.
     * Assumes that supplied event names are names of events that could be
     * generated by this bean.
     * 
     * @param eventName
     *            the name of the event in question
     * @return true if the named event could be generated at this point in time
     */
    public boolean eventGeneratable(String eventName) {
        if (m_listenee == null) {
            return false;
        }

        if (m_listenee instanceof EventConstraints) {
            if (!((EventConstraints) m_listenee).eventGeneratable("dataSet")) {
                return false;
            }
        }
        return true;
    }
}
